

import math

import numpy as np


def count_equals(v1, v2):
    """Count the number of elements that are exactly equals.

    Args:
        v1 (numpy.ndarray): Vector to compare.
        v2 (numpy.ndarray): Vector to compare.

    Returns:
        Number of equal elements.

    """
    if v1.ndim != 1 or v2.ndim != 1:
    	raise ValueError("count_equals expect two vectors")
    return np.count_nonzero(v1 == v2)


def get_f(expr):
    """Returns a 2-dimensional function that evaluates the expression
    passed by parameter.

    Args:
        expr (str): String with a mathematical expression that uses "x"
                    and "y" variables. Must be a valid Python expression
                    because it will be evaluated using "eval" function.
                    The expression can contain also calls to functions in
                    the Python "math" module (e.g. "sqrt(x)+log(y)"). Be
                    aware that using functions like "sqrt(x)" or "x**0.5"
                    in negative numbers will rise an exception. Change the
                    evaluation domain to positive numbers to avoid that
                    problem.

    Returns:
        Return a lambda function that receives 2 parameters: x and y.
        When the returned function is called the expression is evaluated
        using the given values of x and y.

    """
    return lambda x, y: eval(expr, math.__dict__, {"x": x, "y": y})


def central_difference(f, x, y, h=0.1):
    """Compute the derivative of the function f in x, y using the
    central difference method.

    Args:
        f (function): Function with two parameters (x and y).
        x (float): Coordinate x where the gradient will be computed.
        y (float): Coordinate y where the gradient will be computed.
        h (float): Step size.

    Returns:
        gradient (numpy.ndarray): Two dimensional gradient vector.

    """
    gradient = np.zeros(2)
    gradient[0] = (f(x + h, y) - f(x - h, y)) / (2 * h)
    gradient[1] = (f(x, y + h) - f(x, y - h)) / (2 * h)
    return gradient





import numpy as np


class DataSet():
    """Generic collection of inputs and outputs.

    """

    X = np.empty((0, 0))

    Y = np.empty((0, 0))


    def get_N(self):
        """Gets the number of samples in the dataset.

        """
        # The next two expressions are not necessarily equivalent:
        # self.X.shape[0] and self.Y.shape[0]
        # self.Y.shape[0] <-- Can be 0 if no output assigned.
        return self.X.shape[0]

    def get_d(self):
        """Gets the dimension of each sample in the dataset.

        """
        return self.X.shape[1]

    def get_k(self):
        """Gets the number of outputs of each sample.
        
        """
        return self.Y.shape[1]


def load(filename, delimiter="", n_outputs=1, one_shot_output=False, header=False):
    # Set delimiters if filename has a know extension.
    if delimiter is "":
        if filename.endswith(".csv"):
            delimiter = ","
        else:
            delimiter = None
    # Open file and load dataset from stream.
    return load_from_stream(open(filename), delimiter=delimiter, n_outputs=n_outputs,
                            one_shot_output=one_shot_output, header=header)


def load_from_stream(stream, delimiter=",", n_outputs=1,
                     one_shot_output=False, header=False):
    # Check parameters.
    assert not (one_shot_output and abs(n_outputs) != 1), \
        "If one-shot output is selected the number of outputs must be 1."
    # Read stream.
    data = np.loadtxt(stream, delimiter=delimiter, skiprows=int(header))
    # Check feature dimensions.
    d = data.shape[1]
    assert d >= abs(n_outputs), \
        "Number of outputs greater than number of data columns."
    # Set starts/ends of the submatrixes X and Y.
    if n_outputs <= 0:
        start_X = 0
        end_X = start_Y = d + n_outputs
        end_Y = d
    else:
        start_Y = 0
        end_Y = start_X = n_outputs
        end_X = d
    # Create DataSet object.
    dataset = DataSet()
    dataset.X = data[:, start_X:end_X]
    dataset.Y = data[:, start_Y:end_Y]
    if one_shot_output:
        max_output = dataset.Y.max()
        min_output = dataset.Y.min()
        N = dataset.get_N()
        k = max_output - min_output + 1
        indexes = np.add(dataset.Y, -min_output)
        indexes = indexes.astype(int).reshape(N)
        dataset.Y = np.zeros((N, k))
        dataset.Y[np.arange(0, N), indexes] = 1

    return dataset


def save(file, dataset, delimiter=",", header="", footer=""):
    data = np.column_stack((dataset.Y, dataset.X))
    np.savetxt(file, data, delimiter=delimiter, header=header, footer=footer)




"""
Plotting module.
"""

import numpy as np

import matplotlib.cm as cm
import matplotlib.pyplot as plt
from matplotlib.colors import hex2color
from matplotlib.colors import rgb2hex
from matplotlib.colors import LinearSegmentedColormap


theme = "set1"

classes_colors = {
    "set1" : np.array(["#e41a1c", "#377eb8",
                       "#4daf4a", "#984ea3",
                       "#ff7f00", "#ffff33",
                       "#a65628", "#f781bf",
                       "#999999", "#000000"]),
    "set2" : ["#66c2a5", "#fc8d62",
              "#8da0cb", "#e78ac3",
              "#a6d854", "#ffd92f",
              "#e5c494", "#b3b3b3"],
}

classes_markers = {
    "set1" : np.array(['o'] * 10),
    "set2" : np.array(['o'] * 10),
}

rgb_colors = {
    "set1" : np.array([
        (0.8941176470588236, 0.10196078431372549, 0.10980392156862745),
        (0.21568627450980393, 0.49411764705882355, 0.7215686274509804),
        (0.30196078431372547, 0.6862745098039216, 0.2901960784313726),
        (0.596078431372549, 0.3058823529411765, 0.6392156862745098),
        (1.0, 0.4980392156862745, 0.0),
        (1.0, 1.0, 0.2),
        (0.6509803921568628, 0.33725490196078434, 0.1568627450980392),
        (0.9686274509803922, 0.5058823529411764, 0.7490196078431373),
        (0.6, 0.6, 0.6),
        (0.0, 0.0, 0.0)]),
    "set2" : np.array([]),
}


cdict  = {'red':   ((0.0, 1.0, 1.0),
                   (0.5, 1.0, 1.0),
                   (1.0, 0.0, 0.0)),

         'green': ((0.0, 0.0, 0.0),
                   (0.5, 1.0, 1.0),
                   (1.0, 0.0, 0.0)),

         'blue':  ((0.0, 0.0, 0.0),
                   (0.5, 1.0, 1.0),
                   (1.0, 1.0, 1.0)),
                   
         #'alpha': ((0.0, 0.5, 0.5),
         #          (1.0, 0.5, 0.5))
        }
blue_red = LinearSegmentedColormap('BlueRed', cdict)
plt.register_cmap(cmap=blue_red)


def get_class_color(n_class, format="hex"):
    if "hex" in format:
        return classes_colors[theme][n_class]

    return tuple(rgb_colors[theme][n_class])  # rgb format


def light_hex_color(color_hex, light=0.1):
    rgb = hex2color(color_hex)
    light_rgb = [i + light if i + light <= 1 else 1 for i in rgb]
    return rgb2hex(light_rgb)


def predict_area(model, bounds=[-1, 1, -1, 1], samples=50,
                 x_samples=None, y_samples=None):
    """Evaluates a model on a rectangular area.

    Args:
        model (happyml.models.Hypothesis): Model to evaluate.
    """
    # Check parameters.
    if len(bounds) != 4:
        raise ValueError("bounds need 4 values: [xmin, xmax, ymin, ymax]")
    # Create linspaces and matrices with the x and y coordinates.
    x = np.linspace(bounds[0], bounds[1], num=x_samples or samples)
    y = np.linspace(bounds[2], bounds[3], num=y_samples or samples)
    X, Y = np.meshgrid(x, y)
    # Transforms the grids values to a matrix with two columns:
    # the first are the x coordinates, the second are the y coordinates.
    coordinates = np.array([X, Y]).reshape(2, -1).T
    # Use the model to predict an output on each coordinate pair.
    predicted = model.predict(coordinates)
    # Convert the predicted values to a matrix form.
    Z = predicted.reshape(X.shape)  # or Y.shape
    # Return all the usefull data.
    return X, Y, Z

def grid_function(f, bounds=[-1, 1, -1, 1], samples=50,
                  x_samples=None, y_samples=None):
    """Evaluates a function ``f`` on a rectangular area.

    Args:
        f (function): Function to be applied on each sample of the rectangular
            area. The function must receive two matrix by parameter: one with
            x coordinates an another with y coordinates.

    Keyword Args:
        bounds (array): Position and dimension of the rectangular area.
            Indicated by an array of the form [xmin, xmax, ymin, ymax].
        samples (int): number of x and y samples to be used. The more
            samples, the more precision.
        x_samples (int): Use it when you want a different number
            of samples for x axis.
        y_samples (int): Use it when you want a different number
            of samples for y axis.

    Returns:
        X, Y, Z (np.array): 3 matrices of the same size, i.e. of size
        ``(x_samples, y_samples)``.

        **X** (np.array): :math:`X_{ij}` contains the x coordinate of each\
            :math:`Z_{ij}` element. All the rows in a column contains the\
            same number.

        **Y** (np.array): :math:`Y_{ij}` contains the y coordinate of each\
            :math:`Z_{ij}` element. All the columns in a row contains the\
            same number.

        **Z** (np.array): result of appliying ``f`` on :math:`(X_{ij}, Y_{ij})`\
            coordinates. In other words ``Z[i, j] = f(X[i, j], Y[i, j])``.

    Raises:
        ValueError: if ``len`` of ``bounds`` is not 4.

    Example:
        .. code-block:: python

            def fun(X, Y):
                # Linear function. X and Y are matrices, then the output
                # will be a matrix (Z matrix).
                return X * w1 + Y * w2 + b

            # Evaluates function fun on a square centered on the origin
            # with sides of length 2 (from -1 to 1).
            X, Y, Z = grid_function(fun)
    """
    if len(bounds) != 4:
        raise ValueError("bounds need 4 values: [xmin, xmax, ymin, ymax]")
    x = np.linspace(bounds[0], bounds[1], num=x_samples or samples)
    y = np.linspace(bounds[2], bounds[3], num=y_samples or samples)
    X, Y = np.meshgrid(x, y)
    return X, Y, f(X, Y)


def grid_function_slow(f, bounds=[-1, 1, -1, 1], samples=50,
                       x_samples=None, y_samples=None):
    x = np.linspace(bounds[0], bounds[1], num=x_samples or samples)
    y = np.linspace(bounds[2], bounds[3], num=y_samples or samples)
    X, Y = np.meshgrid(x, y)
    Z = np.zeros(shape=X.shape)
    for i in range(Z.shape[0]):
        for j in range(Z.shape[1]):
            Z[i, j] = f(X[i, j], Y[i, j])
    return X, Y, Z


def contourf(fig, f, bounds=[-1, 1, -1, 1], limits=[-1, -0.5, 0, 0.5, 1],
             colors=('#FF6666', '#FF8888', '#8888FF', '#6666FF'),
             x_samples=50, y_samples=50):
    X, Y, Z = grid_function(f, bounds=bounds, x_samples=x_samples, y_samples=y_samples)
    
    return fig.contourf(X, Y, Z, limits, colors=colors, origin='lower', extend='both')


def contour(f, fig=plt, bounds=[-1, 1, -1, 1], limits=[0, 2],
            colors=None, samples=10,
            x_samples=50, y_samples=50):
    X, Y, Z = grid_function(f, bounds=bounds, samples=samples,
                            x_samples=x_samples, y_samples=y_samples)
    
    return fig.contour(X, Y, Z, np.linspace(limits[0], limits[1], num=samples), colors=colors, origin='lower', extend='both')


def heatmap(f, fig=plt, bounds=[-1, 1, -1, 1], limits=[-1, -0.5, 0, 0.5, 1],
             cmap='coolwarm', samples=50, x_samples=None, y_samples=None):
    X, Y, Z = predict_area(f, bounds=bounds, samples=samples,
                           x_samples=x_samples, y_samples=y_samples)
    Z[Z >  1] =  1
    Z[Z < -1] = -1
    return fig.imshow(Z, interpolation='bilinear', origin='lower', cmap=plt.get_cmap(cmap), extent=bounds)


def pcolor(fig, f, bounds=[-1, 1, -1, 1], cmap=cm.coolwarm, samples=50,
           x_samples=None, y_samples=None):
    X, Y, Z = grid_function(f, bounds=bounds, samples=samples,
                            x_samples=x_samples, y_samples=y_samples)
    #return fig.pcolor(X, Y, Z, cmap=cmap, vmin=-1, vmax=+1)
    return fig.pcolormesh(X, Y, Z, cmap=cmap, vmin=-1, vmax=+1)

def plot(dataset, fig=plt, marker='o', s=50, linewidth=0.25):
    """Plot a dataset.

    Warning: this function changes the scale of the plot. Use something
    like::

        ax = plt.gca()
        ax.set_autoscale_on(False)

    to avoid it.

    """
    classes = dataset.Y.flatten().astype(int)
    return fig.scatter(dataset.X[:, 0], dataset.X[:, 1],
                       c=classes_colors[theme][classes], s=s,
                       linewidth=linewidth, marker=marker, zorder=10)

def show():
    plt.show()





import numpy as np


class Hypothesis(object):

    def h(self, x):
        return 0

    def predict(self, X):
        return np.zeros(X.shape[0])


class Perceptron(Hypothesis):

    def __init__(self, w=None, b=None):
        self.w = w if w is not None else np.zeros(2)
        self.b = b if b is not None else 0

    def h(self, x):
        return np.sign(np.dot(self.w.T, x) + self.b)

    def predict(self, X):
        return np.sign(np.dot(X, self.w) + self.b)


class LinearRegression(Hypothesis):

    def __init__(self, w=None, b=None):
        self.w = w if w is not None else np.zeros(2)
        self.b = b if b is not None else 0

    def h(self, x):
        return np.dot(self.w.T, x) + self.b

    def predict(self, X):
        return np.dot(X, self.w) + self.b





"""
Machine Learning library for academic purposes.
GitHub: https://github.com/guiferviz/happyml-py
"""


__version__ = '0.0.3'


def greet():
    """
    HappyML is so friendly that greets you!
    Prints a happy message to the standard output.
    """
    print("Those about to learn we salute you :)")





from os import listdir
from os.path import join

from setuptools import setup

from happyml import __version__


# Install all files under scripts dir.
SCRIPTS_DIR = "scripts"
scripts = [join(SCRIPTS_DIR, i) for i in listdir(SCRIPTS_DIR)]

# Read requirements.
install_requires = [line for line in
            open("requirements.txt").read().splitlines() if line]
tests_require = [line for line in
            open("requirements-tests.txt").read().splitlines() if line]

setup(name='happyml',
      version=__version__,
      description='Machine Learning library for educational purpose.',
      long_description=open('README.rst').read(),
      keywords='happy machine learning',
      url='https://github.com/guiferviz/happyml-py',
      author='guiferviz',
      author_email='guiferviz@gmail.com',
      license='MIT',
      packages=['happyml'],
      install_requires=install_requires,
      tests_require=tests_require,
      test_suite='nose.collector',
      zip_safe=False,
      scripts=scripts)
